/*
Барометр с цифровым и графическим отображение данных атмосферного давления.
Компоненты: Экран текстовый LCD 2004, платформа: arduinio nano/uno (Atmega328 или больше),датчик давления: BME 280.
Amatroskin 2022 год.
*/

#include <LiquidCrystal_I2C.h>
#include <Wire.h>
#include "GyverButton.h"
#include <Adafruit_Sensor.h>                 
#include <Adafruit_BME280.h>                 

#define SEALEVELPRESSURE_HPA (1013.25)       // Задаем высоту
#define BTN_PIN 3                            //Пин кнопки
#define BASE_PERIOD 675000                   //Период набора основного массива 675000мс = 11,25 мин (*16 столбиков = 3 часа) максимальное разрешение графика
#define MIN_VAL 730                          //Минимально значение отображаемое на графике
#define MAX_VAL 760                          //Максимальное значение отображаемое на графике 

LiquidCrystal_I2C lcd(0x27, 20, 4);          //Создаем необходимые объекты
GButton butt1(BTN_PIN);
Adafruit_BME280 bme;

uint32_t tmr1, tmr2;                          //переменные таймеров
uint32_t  set_period = BASE_PERIOD;           //Период отображения, изменяется кратно 3 часам (3, 6, 12, 24)
int16_t plot_array[20];                       //Массив данных для отрисовки графика
uint16_t base_array [128];                    //Базовый массив хранит все измерения за последние сутки (128 ячеек * 11,25 минут = 1440 минут = 24 часа)
int16_t value, delta;                         //Текушие (снимаемые) показания, разница показаний за выбранный отрезок времени
byte interval = 1;                            //Интервал отображаемый на экране (разница во времени между соседними столбиками диаграммы)
                                              // 11,25 мин * 16 = 3 часа, 22,5 мин - 6 часов, 45мин - 12 часов, 90 мин - 24 часа
void setup() {
  read_all ();
  //  Serial.begin(9600);
  attachInterrupt(1, isr, CHANGE);
  butt1.setDebounce(80);                      // настройка антидребезга (по умолчанию 80 мс)
  butt1.setTimeout(300);                      // настройка таймаута на удержание (по умолчанию 500 мс)
  lcd.init();
  lcd.backlight();
  lcd.clear();
  if (!bme.begin(0x76)) {                              //Инициализация датчика BME280 по адресу 0x76 (по умолчанию)
    //Serial.println("Could not find a valid BME280!");//Печать сообщения об ошибке в монитор порта
    lcd.setCursor(3, 1);
    lcd.print(F("No connection"));                     //___________________________или на самом экране
    lcd.setCursor(5, 2);
    lcd.print(F("to sensor"));
    while (1);                                          //без датчика дальше не пойдет
  }

  if (!digitalRead(BTN_PIN)) {                          //Сброс настроек при включении с зажатой кнопкой
    for (byte i = 0; i < 128; i++) base_array [i] = 0;  //Стереть данные
    update_all ();                                      //Сохранить
    lcd.setCursor(5, 1);                                //Сообщить об этом
    lcd.print(F("Reset data"));
    lcd.setCursor(9, 2);
    lcd.print(F("OK"));
  }
  while (!digitalRead(BTN_PIN));
  lcd.clear();
    
  initPlot();                                           //Инициализация символов для отрисовки
  
  value = round ((bme.readPressure() / 133.32239023154F));//Снимаем текущие показания, конвертируем в мм.рт.ст.
  base_array[0] = value;
  get_data ();
}

void isr() {                                              //Опрашиваем кнопку в прерывании, чтобы поймать нажатие в любом случае
  butt1.tick();
}

void loop() {
  butt1.tick();                                           //Опрос кнопки
  if (butt1.isClick()) {                                  //При нажатии:
    interval *= 2;                                        //Переключаем масштаб диаграммы
    if (interval > 8) interval = 1;
    set_period = BASE_PERIOD * interval;                  //Пересчитываем интервал обновления данных с датчика
    get_data ();                                          //Обновляем информацию на дисплее
  }
                            
  if (millis() - tmr1 >= BASE_PERIOD) {                   //Набираем базовый массив данных
    tmr1 = millis();                                      //Каждые 11,25 минут снимаем показания с датчика
    for (int i = 126; i >= 0; i--) {                      //Сдвигаем весь массив на один пункт
      base_array[i + 1] = base_array[i];
    }
    value = round ((bme.readPressure() / 133.32239023154F));//Переводим в мм/рт.ст.
    base_array[0] = value;                                 //Записываем последние показания в массив
    update_all ();                                         //И запоминаем в EEPROM
  }

  if (millis() - tmr2 >= set_period) {                     //Через промежуток времени set_period (миллисекунды)
    tmr2 = millis();                                       //Обновляем данные на экране
    get_data ();
  }
}

void get_data () {                                          //Отрисовываем 16 столбиков диагаммы на экране
  for (int i = 15; i >= 0; i--) {
    drawPlot(0, 3, 16, 4, MIN_VAL, MAX_VAL, (base_array[i * interval]));
  }
  delta = ((base_array[0]) - (base_array[15 * interval]));  //Расчитываем дельту (изменение давления) за выбранный интервал 
  screen_data (value, delta, (interval * 3));               //Выводим текстовую информацию на экран
}

void screen_data (int value, int delta, byte interval) {    //Функция вывода текстовой информациина экран (все кроме столбиков диаграмы) 
  lcd.setCursor(17, 0);
  lcd.print(value);
  lcd.setCursor(17, 2);
  if (delta == value) delta = 0;
  if (delta > 0) {
    lcd.print("+");
  } else if (delta < 0) {
    lcd.print("-");
  } else if (delta == 0) {
    lcd.print(" ");
  }
  lcd.setCursor(18, 2);
  lcd.print(abs(delta));
  if (abs(delta) < 10) {
    lcd.setCursor(19, 2);
    lcd.print(" ");
  }
  lcd.setCursor(17, 1);
  lcd.print("mmH");
  lcd.setCursor(17, 3);
  lcd.print(interval);
  (interval < 10) ? lcd.print("h ") : lcd.print("h");
}

void initPlot() {
  // необходимые символы для работы
  // создано в http://maxpromer.github.io/LCD-Character-Creator/
  byte row8[8] = {0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row7[8] = {0b00000,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row6[8] = {0b00000,  0b00000,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row5[8] = {0b00000,  0b00000,  0b00000,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row4[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row3[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b11111,  0b11111,  0b11111};
  byte row2[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b11111,  0b11111};
  byte row1[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b11111};
  lcd.createChar(0, row8);
  lcd.createChar(1, row1);
  lcd.createChar(2, row2);
  lcd.createChar(3, row3);
  lcd.createChar(4, row4);
  lcd.createChar(5, row5);
  lcd.createChar(6, row6);
  lcd.createChar(7, row7);
}

                                              //Отрисовка диаграммы, код взят у Гайвера https://alexgyver.ru/lcd-plots-and-bars/
void drawPlot(byte pos, byte row, byte width, byte height, int min_val, int max_val, int fill_val) {

  for (byte i = 0; i < width; i++) {
    plot_array[i] = plot_array[i + 1];
  }
  fill_val = constrain(fill_val, min_val, max_val);
  plot_array[width - 1] = fill_val;
  for (byte i = 0; i < width; i++) {         // каждый столбец параметров
    int infill, fract;
    
                                             // найти количество целых блоков с учётом минимума и максимума для отображения на графике
    infill = floor((float)(plot_array[i] - min_val) / (max_val - min_val) * height * 10);
    fract = (infill % 10) * 8 / 10;                          // найти количество оставшихся полосок
    infill = infill / 10;
    for (byte n = 0; n < height; n++) {     // для всех строк графика
      if (n < infill && infill > 0) {       // пока мы ниже уровня
        lcd.setCursor(i, (row - n));        // заполняем полными ячейками
        lcd.write(0);
      }
      if (n >= infill) {                          // если достигли уровня
        lcd.setCursor(i, (row - n));
        if (fract > 0) lcd.write(fract);          // заполняем дробные ячейки
        else lcd.write(16);                       // если дробные == 0, заливаем пустой
        for (byte k = n + 1; k < height; k++) {   // всё что сверху заливаем пустыми
          lcd.setCursor(i, (row - k));
          lcd.write(16);
        }
        break;
      }
    }
  }
}

void update_all () {                             //Обновляем данные в EEPROM
  eeprom_update_block((void*)&base_array, 0, sizeof(base_array));
}

void read_all () {                              //Читаем данные из EEPROM
  eeprom_read_block((void*)&base_array, 0, sizeof(base_array));
}
